// Copyright (C) 2017 Google Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package app

import (
	"flag"
	"fmt"
	"os"
	"sync"
	"syscall"
	"time"

	"github.com/google/gapid/core/app/analytics"
	"github.com/google/gapid/core/app/crash"
	"github.com/google/gapid/core/app/crash/reporting"
	"github.com/google/gapid/core/app/status"
	"github.com/google/gapid/core/event/task"
	"github.com/google/gapid/core/fault"
	"github.com/google/gapid/core/log"
	"github.com/google/gapid/core/os/file"
	"github.com/google/gapid/core/text"
	"github.com/pkg/errors"
)

var (
	// Name is the full name of the application
	Name string
	// Flags is the main application flags.
	Flags AppFlags
	// ExitFuncForTesting can be set to change the behaviour when there is a command line parsing failure.
	// It defaults to os.Exit
	ExitFuncForTesting = os.Exit
	// ShortHelp should be set to add a help message to the usage text.
	ShortHelp = ""
	// ShortUsage is usage text for the additional non-flag arguments.
	ShortUsage = ""
	// UsageFooter is printed at the bottom of the usage text
	UsageFooter = ""
	// Version holds the version specification for the application.
	// The default version is the one defined in version.cmake.
	// If valid a command line option to report it will be added automatically.
	Version VersionSpec
	// Restart is the error to return to cause the app to restart itself.
	Restart = fault.Const("Restart")
)

// VersionSpec is the structure for the version of an application.
type VersionSpec struct {
	// Major version, the version structure is in valid if <0
	Major int
	// Minor version, not used if <0
	Minor int
	// Point version, not used if <0
	Point int
	// The build identifier, not used if an empty string
	Build string
}

// IsValid reports true if the VersionSpec is valid, ie it has a Major version.
func (v VersionSpec) IsValid() bool {
	return v.Major >= 0
}

// GreaterThan returns true if v is greater than o.
func (v VersionSpec) GreaterThan(o VersionSpec) bool {
	switch {
	case v.Major > o.Major:
		return true
	case v.Major < o.Major:
		return false
	case v.Minor > o.Minor:
		return true
	case v.Minor < o.Minor:
		return false
	case v.Point > o.Point:
		return true
	default:
		return false
	}
}

func (v VersionSpec) String() string {
	return fmt.Sprintf("%v", v)
}

// Format implements fmt.Formatter to print the version.
func (v VersionSpec) Format(f fmt.State, c rune) {
	fmt.Fprint(f, v.Major)
	if v.Minor >= 0 {
		fmt.Fprint(f, ".", v.Minor)
	}
	if v.Point >= 0 {
		fmt.Fprint(f, ".", v.Point)
	}
	if v.Build != "" {
		fmt.Fprint(f, ":", v.Build)
	}
}

func init() {
	Name = file.Abs(os.Args[0]).Basename()
	Flags.Log = logDefaults()
}

// getArgs preprocesses the command line arguments to
// split out any packed arguments passed in as a string.
func getArgs() []string {
	// parse the command line
	args := os.Args[1:]
	for {
		expanded := false
		newArgs := []string{}
		for i := 0; i < len(args); i++ {
			arg := args[i]
			if (arg == "-args" ||
				arg == "--args") &&
				i != len(args)-1 {
				newArgs = append(newArgs, text.SplitArgs(args[i+1])...)
				i++
				expanded = true
			} else {
				newArgs = append(newArgs, arg)
			}
		}
		args = newArgs
		if !expanded {
			return args
		}
	}
}

// Run performs all the work needed to start up an application.
// It parsers the main command line arguments, builds a primary context that will be cancelled on exit
// runs the provided task, cancels the primary context and then waits for either the maximum shutdown delay
// or all registered signals whichever comes first.
func Run(main task.Task) {
	// Defer the panic handling
	defer func() {
		switch cause := recover().(type) {
		case nil:
		case ExitCode:
			ExitFuncForTesting(int(cause))
		default:
			crash.Crash(cause)
		}
	}()

	// If run via 'bazel run', use the shell's CWD, not bazel's.
	if cwd := os.Getenv("BUILD_WORKING_DIRECTORY"); cwd != "" {
		os.Chdir(cwd)
	}

	// install all the common application flags
	rootCtx := prepareContext(&Flags.Log)

	args := getArgs()

	flag.CommandLine.Usage = func() { Usage(rootCtx, "") }
	verbMainPrepare(&Flags)
	globalVerbs.flags.Parse(nil, args...)

	// Force the global verb's flags back into the default location for
	// main programs that still look in flag.Args()
	// TODO: We need to stop doing this
	globalVerbs.flags.ForceCommandLine()

	if Flags.FullHelp {
		Usage(rootCtx, "")
		return
	}

	if Flags.DecodeStack != "" {
		stack := decodeCrashCode(Flags.DecodeStack)
		fmt.Fprintf(os.Stdout, "Stack dump:\n%v\n", stack)
		return
	}

	if Flags.Version {
		fmt.Fprint(os.Stdout, Name, " version ", Version, "\n")
		return
	}

	endProfile := applyProfiler(rootCtx, &Flags.Profile)

	ctx, cancel := task.WithCancel(rootCtx)

	ctx = updateContext(ctx, &Flags.Log)

	if Flags.Analytics != "" {
		analytics.Enable(ctx, Flags.Analytics, analytics.AppVersion{
			Name: Name, Build: Version.Build,
			Major: Version.Major, Minor: Version.Minor, Point: Version.Point,
		})
		analytics.SendEvent("app", "start", Name)
	}

	if Flags.CrashReport {
		reporting.Enable(ctx, Name, Version.String())
	}

	if Flags.Log.Status {
		status.RegisterLogger(time.Second)
	}

	if Flags.Profile.Trace != "" {
		f, err := os.Create(Flags.Profile.Trace)
		if err != nil {
			log.E(ctx, "Could not start trace profiling")
		} else {
			status.RegisterTracer(f)
			defer f.Close()
		}
	}

	// Defer the shutdown code
	shutdownOnce := sync.Once{}
	shutdown := func() {
		shutdownOnce.Do(func() {
			analytics.Flush()
			LogHandler.Close()
			cancel()
			if !WaitForCleanup(rootCtx) {
				fmt.Fprint(os.Stderr, "Timeout waiting for cleanup")
			}
			endProfile()
		})
	}

	defer shutdown()

	// Add the abort and crash signal handlers
	handleAbortSignals(shutdown)
	handleCrashSignals(ctx, shutdown)

	// Now we are ready to run the main task
	err := main(ctx)
	if errors.Cause(err) == Restart {
		err = doRestart()
	}
	if err != nil {
		log.E(ctx, "Main failed\nError: %v", err)
	}
}

func doRestart() error {
	argv0 := file.ExecutablePath().System()
	files := make([]*os.File, syscall.Stderr+1)
	files[syscall.Stdin] = os.Stdin
	files[syscall.Stdout] = os.Stdout
	files[syscall.Stderr] = os.Stderr
	wd, err := os.Getwd()
	if nil != err {
		return err
	}
	_, err = os.StartProcess(argv0, os.Args, &os.ProcAttr{
		Dir:   wd,
		Env:   os.Environ(),
		Files: files,
		Sys:   &syscall.SysProcAttr{},
	})
	return err
}
